﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Reflection;
using System.Reflection.Emit;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;
using MetadataLib.Interop;

namespace MetadataLib
{
    public static class MetadataUtil
    {
        private const uint mdTokenNil = 0;

        private const uint E_NOINTERFACE = 0x80004002;

        internal const string IID_IMetaDataImport = "7DAC8207-D3AE-4C75-9B67-92801A497D44";
        
        internal const string IID_IMetaDataAssemblyImport = "EE62470B-E94B-424E-9B7C-2F00C9249F93";

        [MethodImpl(MethodImplOptions.NoInlining)]
        public static MethodInfo GetSetterName(Action del)
        {
            MethodInfo mi = null;

            var assembly = Assembly.GetCallingAssembly();

            var delMethod = del.Method;
            var originMvid = delMethod.Module.ModuleVersionId;
            var il = delMethod.GetMethodBody().GetILAsByteArray();

            uint token;
            for (var pos = il.Length - 1; pos >= 0; pos--)
            {
                token = mdTokenNil;
                var cur = il[pos];

                if ((cur == OpCodes.Call.Value) || (cur == OpCodes.Callvirt.Value))
                {
                    token = TryGetToken(pos, il);
                    if (token != mdTokenNil)
                    {
                        var tokenType = GetTokenType(token);
                        var isMethodDef = tokenType.Equals(MetadataTokenType.MethodDef);
                        var isMemberRef = tokenType.Equals(MetadataTokenType.MemberRef);

                        if (isMethodDef || isMemberRef)
                        {
                            mi = Token2MethodInfo(token, assembly, isMethodDef, originMvid);
                            break;
                        }
                    }
                }
            }

            return mi;
        }

        private static uint TryGetToken(int position, byte[] il)
        {
            uint token = mdTokenNil;

            const int tokenSize = 4;

            if ((position + tokenSize) < il.Length)
            {
                token = (uint)(il[position + 4] << 24) + (uint)(il[position + 3] << 16) + (uint)(il[position + 2] << 8) + (uint)il[position + 1];
            }

            return token;
        }

        private static IMetaDataImport GetImport(Assembly assemblyScope)
        {
            var dispenser = new CorMetaDataDispenserExClass();

            var dispenserEx = dispenser as IMetaDataDispenserEx;

            if (dispenserEx == null)
            {
                Check(E_NOINTERFACE);
            }
            
            Guid metaDataImportGuid = new Guid(IID_IMetaDataImport);

            object rawScope = null;
            var hr = dispenserEx.OpenScope(assemblyScope.Location, 0, ref metaDataImportGuid, out rawScope);

            Check(hr);

            var import = rawScope as IMetaDataImport;

            if (import == null)
            {
                Check(E_NOINTERFACE);
            }

            return import;
        }

        private static MethodInfo Token2MethodInfo(uint token, Assembly assemblyScope, bool isMethodDef, Guid originMvid)
        {
            MethodInfo mi = null;

            uint methodDef;
            Guid mvid;
            Assembly assembly;

            if (isMethodDef)
            {
                methodDef = token;
                assembly = assemblyScope;
                mvid = originMvid;
            }
            else
            {
                ResolveNewScope(token, assemblyScope, out methodDef, out assembly, out mvid);
            }

            mi = GetMethodInfoFromScope(methodDef, assembly, mvid); 

            return mi;
        }

        private static void ResolveNewScope(uint token, Assembly assemblyScope, out uint newToken, out Assembly newAssembly, out Guid newMvid)
        {
            newToken = mdTokenNil;
            newAssembly = null;
            newMvid = Guid.Empty;

            var import = GetImport(assemblyScope);

            uint typeRef;
            uint memberRefNameLen = 0;
            uint memberRefSigLen;
            var memberRefName = new char[1024];
            IntPtr ptrMemberRefSig;

            var hr = import.GetMemberRefProps(token, out typeRef, memberRefName, (uint)memberRefName.Length, out memberRefNameLen, out ptrMemberRefSig, out memberRefSigLen);

            Check(hr);

            uint typeDef = mdTokenNil;
            var newImport = ResolveNewImport(import, typeRef, out typeDef);

            var name = new string(memberRefName, 0, (int)memberRefNameLen - 1);
            var methodDef = ResolveToken(newImport, typeDef, name);

            uint len;
            Guid mvid = Guid.Empty;
            hr = newImport.GetScopeProps(null, 0, out len, ref mvid);

            Check(hr);

            var asmName = ResolveAssemblyName(newImport);

            newToken = methodDef;
            newAssembly = AppDomain.CurrentDomain.GetAssemblies().FirstOrDefault(a => a.GetName().FullName.Equals(asmName.FullName));
            if (newAssembly == null)
            {
                newAssembly = Assembly.Load(asmName);
            }
            newMvid = mvid;
        }

        private static uint ResolveToken(IMetaDataImport import, uint typeDef, string name)
        {
            uint methodDef = mdTokenNil;

            var hEnum = IntPtr.Zero;
            var methodTokens = new uint[1024];
            var methodName = new char[1024];
            uint len;

            try
            {
                do
                {
                    var hr = import.EnumMethods(ref hEnum, typeDef, methodTokens, (uint)methodTokens.Length, out len);

                    Check(hr);

                    for (var i = 0; i < len; i++)
                    {
                        uint tmpTypeDef;
                        uint attrs;
                        var sig = IntPtr.Zero;
                        uint sigLen;
                        uint rva;
                        uint flags;
                        uint methodNameLen;
                        hr = import.GetMethodProps(methodTokens[i], out tmpTypeDef, methodName, (uint)methodName.Length, out methodNameLen, out attrs, out sig, out sigLen, out rva, out flags);

                        Check(hr);

                        var curName = new string(methodName, 0, (int)methodNameLen - 1);

                        if (name == curName)
                        {
                            methodDef = methodTokens[i];
                            break;
                        }
                    }

                    if (len < methodTokens.Length)
                    {
                        break;
                    }
                }
                while (methodDef == mdTokenNil);
            }
            finally
            {
                if (hEnum != IntPtr.Zero)
                {
                    import.CloseEnum(hEnum);
                }
            }

            return methodDef;
        }

        private static IMetaDataImport ResolveNewImport(IMetaDataImport import, uint typeRef, out uint typeDef)
        {
            typeDef = mdTokenNil;
            IMetaDataImport newImport = null;

            var parents = new Stack<uint>();

            while (GetTokenType(typeRef) == MetadataTokenType.TypeRef)
            {
                parents.Push(typeRef);
                
                uint resScope;
                uint parentNameLen;
                var hr = import.GetTypeRefProps(typeRef, out resScope, null, 0, out parentNameLen);

                Check(hr);

                typeRef = resScope;
            }

            object newRawScope = null;

            Guid metaDataImportGuid = new Guid(IID_IMetaDataImport);

            while (parents.Count > 0)
            {
                typeRef = parents.Pop();

                var hr = import.ResolveTypeRef(typeRef, ref metaDataImportGuid, out newRawScope, out typeDef);

                Check(hr);
            }

            newImport = newRawScope as IMetaDataImport;

            if (newImport == null)
            {
                Check(E_NOINTERFACE);
            }

            return newImport;
        }

        private static AssemblyName ResolveAssemblyName(IMetaDataImport import)
        {
            var asmImport = import as IMetaDataAssemblyImport;

            if (asmImport == null)
            {
                Check(E_NOINTERFACE);
            }

            uint assemblyId;
            var hr = asmImport.GetAssemblyFromScope(out assemblyId);

            Check(hr);

            IntPtr ptrPublicKey;
            uint lenPublicKey;
            uint hashAlgId;
            var assemblyName = new char[1024];
            uint assemblyNameLen;
            ASSEMBLYMETADATA asmMetaData;
            uint asmFlags;
            hr = asmImport.GetAssemblyProps(assemblyId, out ptrPublicKey, out lenPublicKey, out hashAlgId, assemblyName, (uint)assemblyName.Length, out assemblyNameLen, out asmMetaData, out asmFlags);

            Check(hr);

            var asmName = new AssemblyName();
            asmName.Name = new string(assemblyName, 0, (int)assemblyNameLen - 1);
            asmName.Version = new Version(asmMetaData.usMajorVersion, asmMetaData.usMinorVersion, asmMetaData.usBuildNumber, asmMetaData.usRevisionNumber);
            if (lenPublicKey > 0)
            {
                var pkey = new byte[lenPublicKey];
                Marshal.Copy(ptrPublicKey, pkey, 0, pkey.Length);
                asmName.SetPublicKey(pkey);
            }
            asmName.CultureInfo = asmMetaData.cbLocale == 0
                                ? CultureInfo.InvariantCulture
                                : CultureInfo.GetCultureInfo((int)asmMetaData.cbLocale);

            return asmName;
        }

        private static MethodInfo GetMethodInfoFromScope(uint token, Assembly assemblyScope, Guid mvid)
        {
            MethodInfo mi = null;

            if ((token != mdTokenNil) && (assemblyScope != null) && (mvid != Guid.Empty))
            {
                const BindingFlags flags = BindingFlags.Static | BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic;

                var q = from module in assemblyScope.GetModules()
                        from type in module.GetTypes()
                        from method in type.GetMethods(flags)
                        where (method.MetadataToken == token) && (method.Module.ModuleVersionId == mvid)
                        select method;

                mi = q.SingleOrDefault();
            }

            return mi;
        }

        private static MetadataTokenType GetTokenType(uint token)
        {
            return (MetadataTokenType)(token & 0xFF000000);
        }

        private static void Check(uint hr)
        {
            if (hr != 0)
            {
                Marshal.ThrowExceptionForHR((int)hr);
            }
        }
    }
}
